/*
 * module.hpp
 *
 *  Created on: 20/02/2013
 *      Author: mmcgill
 */

#ifndef CROSBOT_FASTSLAM_MODULE_HPP_
#define CROSBOT_FASTSLAM_MODULE_HPP_

#include <ros/ros.h>

#include <crosbot/controls/command.hpp>
#include <crosbot_map/ListSnaps.h>
#include <crosbot_map/LoadSnaps.h>
#include <crosbot_fastslam/fastslam.hpp>
#include <geometry_msgs/Pose.h>
#include <nav_msgs/Path.h>
#include <nav_msgs/SetMap.h>
#include <sensor_msgs/LaserScan.h>
#include <tf/transform_listener.h>
#include <tf/transform_broadcaster.h>

namespace crosbot {

namespace fastslam {

#define DEFAULT_ODOMFRAME			"/icp"
#define DEFAULT_MAPFRAME			"/map"
#define DEFAULT_BASEFRAME			"/base_link"
#define DEFAULT_MAXWAIT4TRANSFORM	2.0			// [s]

class FastSLAMModule;
typedef Handle<FastSLAMModule> FastSLAMModulePtr;

class FastSLAMModule : public FastSLAMMap, virtual public MapListener {
protected:
	// ROS Configuration
    std::string baseFrame;
    std::string mapFrame;
	std::string odomFrame;
	std::string pub_historyName;
	std::string pub_mapName;
	std::string pub_particlesName;
	std::string srv_listSnapsName;
	std::string srv_loadSnapsName;
	std::string srv_setMapName;
	std::string sub_poseEstimate;
	std::string sub_scanName;
	std::string sub_snapName;

	ros::ServiceServer listSnaps_srv;
	ros::ServiceServer loadSnaps_srv;
	ros::ServiceServer setMap_srv;
	ros::Subscriber poseEstimateSub;
	ros::Subscriber scanSub;
	ros::Subscriber snapSub;

	ros::Publisher gridPub;
	ros::Publisher historyPub;
	ros::Publisher particlesPub;

	crosbot::CrosbotCommandPtr crosbotCommand;

	tf::TransformListener tfListener;
    tf::TransformBroadcaster tfBroadcast;

public:

    FastSLAMModule(std::string name = "");
    virtual ~FastSLAMModule();

    void initialize(ros::NodeHandle& nh);
    virtual void start();
    virtual void stop();

	// ROS Callbacks
    void callbackPoseEstimate(const geometry_msgs::PointStampedConstPtr& poseEstimate);
    void callbackScan(const sensor_msgs::LaserScanConstPtr& latestScan);
	void callbackSnap(const crosbot_map::SnapMsgConstPtr& snap);
	bool service_listSnaps(crosbot_map::ListSnapsRequest &request, crosbot_map::ListSnapsResponse &response);
	bool service_loadSnaps(crosbot_map::LoadSnapsRequest &request, crosbot_map::LoadSnapsResponse &response);
	bool service_setMap(nav_msgs::SetMapRequest &req, nav_msgs::SetMapResponse &res);

	void publishTransform(const Pose& robot, const Pose& odometry, const Time& time);

	// Map Listener callbacks
	void motionTracked(MapPtr map);
	void mapUpdated(MapPtr map);
	void tagAdded(MapPtr map, TagPtr tag);
	void tagChanged(MapPtr map, TagPtr tag);

	// Crosbot command callback
	void callback_receivedCommand(crosbot_msgs::ControlCommandPtr command);

protected:

    class QueuedAction_Module : public QueuedAction {
    public:
        nav_msgs::SetMapRequestPtr setMap;

        virtual ~QueuedAction_Module() {};
    };
    std::shared_ptr<QueuedAction_Module> QueuedAction_ModulePtr;

	virtual void resetMap_action();
	virtual void loadMap_action(QueuedActionPtr action);
	virtual void estimatePose_action(QueuedActionPtr action);

private:

	/**
	 * Estimates the pose of the robot within the entire map.
	 * Does not work well.
	 */
	void estimatePoseInMap()  __attribute__ ((deprecated));
};

} // namespace fastslam

} // namespace crosbot

#endif /* CROSBOT_FASTSLAM_MODULE_HPP_ */
